;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "consult-dir" "20251020.416"
  "Insert paths into the minibuffer prompt."
  '((emacs   "26.1")
    (project "0.3.0")
    (consult "2.0"))
  :url "https://github.com/karthink/consult-dir"
  :commit "1497b46d6f48da2d884296a1297e5ace1e050eb5"
  :revdesc "1497b46d6f48"
  :keywords '("convenience")
  :authors '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com"))
  :maintainers '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")))
